/***********************************************************************/
/*                                                                     */
/* MODULE NAME        DDESERV2.C                                       */
/*                                                                     */
/* DESCRIPTIVE NAME   DDE Server for CICS OS/2 ECI                     */
/*                                                                     */
/* Statement:         Sample Code - No Warranty Provided               */
/*                                                                     */
/*                    (c) Copyright IBM Corp. 1994.                    */
/*                                                                     */
/*                    All rights reserved.                             */
/*                                                                     */
/*                                                                     */
/* Status:            Version 1 Release 0 Modification 1               */
/*                                                                     */
/*  NOTES :-                                                           */
/*    DEPENDENCIES = OS/2 V2.X                                         */
/*    RESTRICTIONS = none                                              */
/*    MODULE TYPE  = Source Module                                     */
/*    PROCESSOR    = PS/2 and PC                                       */
/*                                                                     */
/* Changes:                                                            */
/*   06/01/95 Modified default Window position to fit VGA screen-KJT1  */
/*                                                                     */
/*                                                                     */
/***********************************************************************/
/*                                                                     */
/***********************************************************************/
/* PROGRAM NAME: DDESERV2                                              */
/*                                                                     */
/* Sample of DDE Server for CICS OS/2 ECI                              */
/* Compilable for C Set/2.                                             */
/*                                                                     */
/***********************************************************************/

#define INCL_PM
#define INCL_WIN
#define INCL_WINDDE
#define INCL_GPI
#define INCL_DOS
#define INCL_ERRORS

#include <os2.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include "faaecih.h"
#include "DDESERV2.h"

/*************************************************************************/
/*   Main Procedure                                                      */
/*************************************************************************/

int main ( )
{
  /* globals for main program */

  HAB         hab ;
  HMQ         hmq ;
  QMSG        qmsg ;
  ULONG       flCreate ;
  PID         pid ;
  TID         tid ;
  HSWITCH     hswitch ;
  SWCNTRL     swctl ;

/***************************************************************************/
/* Initialize PM and create a window with a listbox as a subclassed client */
/* area:                                                                   */
/***************************************************************************/

  hab = WinInitialize ( 0L ) ;
  hmq = WinCreateMsgQueue ( hab, 0 ) ;

  flCreate = FCF_TITLEBAR | FCF_SYSMENU | FCF_MINMAX | FCF_SIZEBORDER |
             FCF_ICON | FCF_ACCELTABLE | FCF_MENU ;

  /* Create the parent window */

  hwndFrame = WinCreateStdWindow ( HWND_DESKTOP,
                                   FS_SCREENALIGN,
                                   &flCreate,
                                   WC_LISTBOX,
                                   "DDE Server for CICS OS/2 ECI",
                                   LS_HORZSCROLL,
                                   0L,
                                   ID_MAINWND,
                                   ( PHWND ) &hwndClient ) ;

  WinSetWindowPos ( hwndFrame, HWND_TOP,
                    10, 10, 500, 300,                              /* KJT1 */
                    SWP_SIZE | SWP_MOVE | SWP_ACTIVATE | SWP_SHOW ) ;

  SystemListboxWindowProc = WinSubclassWindow ( hwndClient,
                                                MainWindowProc ) ;


  /*************************************************************************/
  /* Add the main window to the switch list:                               */
  /*************************************************************************/

  WinQueryWindowProcess ( hwndFrame, &pid, &tid ) ;
  swctl.hwnd = hwndFrame ;
  swctl.hwndIcon = 0L ;
  swctl.hprog = 0L ;
  swctl.idProcess = pid ;
  swctl.idSession = 0L ;
  swctl.uchVisibility = SWL_VISIBLE ;
  swctl.fbJump = SWL_JUMPABLE ;
  sprintf ( swctl.szSwtitle, "DDE Server for CICS OS/2 ECI" ) ;

  hswitch = WinAddSwitchEntry ( &swctl ) ;

  /***************************************************************************/
  /* Register DDEServer Window class for child window to handle each DDE Conv*/
  /***************************************************************************/

  WinRegisterClass(hab, "DDEServerClass", DDEWindowProc, 0L, 0) ;

  /*************************************************************************/
  /* Main Message loop:                                                    */
  /*************************************************************************/

  while ( WinGetMsg ( hab, ( PQMSG ) &qmsg, (HWND)NULL, 0, 0 ) )
    WinDispatchMsg ( hab, ( PQMSG ) &qmsg );

  /***************************************************************************/
  /* Exit:                                                                   */
  /***************************************************************************/

  WinRemoveSwitchEntry ( hswitch ) ;
  WinDestroyWindow ( hwndFrame ) ;
  WinDestroyMsgQueue ( hmq ) ;
  WinTerminate ( hab ) ;

  return 0 ;

}  /* end function main */

/***************************************************************************/
/* Main Window procedure                                                   */
/***************************************************************************/

MRESULT EXPENTRY MainWindowProc ( HWND   hwnd,
                                  ULONG  msg,
                                  MPARAM mp1,
                                  MPARAM mp2 )
{

  /* Global Variables */

  SHORT       sRetCode ;
  USHORT      command;

  /* The main code */

  switch ( msg ) {

  case WM_COMMAND:

    command = SHORT1FROMMP ( mp1 ) ;

    switch ( command ) {

      case ID_CALL_ECI :

        /* set up the commarea to be sent and call the ECI */
        strcpy ( szCommArea, "OS/2 DDE Server for CICS OS/2 ECI Sample" ) ;
        sRetCode = ECICall("ECITEST ", hwndFrame ) ;
        break ;

      case ID_LISTCLEAR :

        WinSendMsg ( hwndClient, LM_DELETEALL, 0, 0L ) ;
        break;

      case ID_EXITPROGRAM :

        WinPostMsg( hwnd, WM_CLOSE, 0L, 0L );
        break;
    }  /* end of switch  command */

  break;   /* end of switch WM_COMMAND  */

  case ECI_MSG :

    /* display the ECI return code */
    Response ( "ECI Async Notify Msg ", SHORT2FROMMP ( mp1 ) );

    if (ECIRequestData() != ECI_NO_ERROR) {
      /* The ECI call has failed. Write an error message to screen */
      strcpy( szCommArea , "ECI Failure, check DDESERV") ;
    } /* end if */

    Info ( "Commarea returned is '%s'",  szCommArea ) ;
    break ;

  case WM_CLOSE :

   /* send DDE_TERMINATE to all connected clients before posting QUIT msg */
    TerminateConversations() ;
    WinPostMsg ( hwnd, WM_QUIT, 0L, 0L ) ;
    break;

  case WM_DDE_INITIATE :

    /**********************************************************************/
    /* This message indicates a client is trying to initiate a DDE        */
    /* conversation with us.                                              */
    /**********************************************************************/
    ServerInitiate((HWND)mp1, mp2) ;
    break;

  default:

    return ( SystemListboxWindowProc ( hwnd, msg, mp1, mp2 ) ) ;

  } /* end switch(msg) */

  return ( MRESULT ) 0 ;

} /* end function MainWindowProc */

/**************************************************************************/
/*                                                                        */
/*  FUNCTION: ECICall                                                     */
/*                                                                        */
/*  PURPOSE:  Call ECI requesting ASYNC_NOTIFY passing commarea           */
/*                                                                        */
/**************************************************************************/

SHORT ECICall (CHAR szProgram[9], HWND hwndToNotify )
{

  int     Size ;
  SHORT   sRetCode ;

  /* for this sample - commarea is limited to 256 bytes */

  Size = sizeof ( szCommArea ) ;
  Info ( "Calling ECI - sending %d bytes. Commarea is '%s'.  Window to Notify is %d.",
         strlen(szCommArea),  szCommArea, hwndToNotify ) ;

  /* clear the ECI control block */
  memset ( &eci_parms, 0, sizeof ( ECI_PARMS ) ) ;

  /* set up the ECI control block for this call */
  eci_parms.eci_call_type                         = ECI_ASYNC_NOTIFY_MSG ;
  memcpy ( &eci_parms.eci_program_name, szProgram , strlen(szProgram)) ;
  memcpy ( &eci_parms.eci_userid      , "SYSAD   ", 8 ) ;
  memcpy ( &eci_parms.eci_password    , "SYSAD   ", 8 ) ;
  eci_parms.eci_commarea                          = szCommArea ;
  eci_parms.eci_commarea_length                   = Size ;
  eci_parms.eci_extend_mode                       = ECI_NO_EXTEND ;
  eci_parms.eci_timeout                           = 20 ;
  eci_parms.eci_async_notify.window_handle        = (FAAHWND) hwndToNotify ;
  eci_parms.eci_message_qualifier                 = 42 ; /* Number chosen at random */
  eci_parms.eci_message_id                        = ECI_MSG ;
  eci_parms.eci_luw_token                         = ECI_LUW_NEW ;

  /*  and call the ECI */
  sRetCode = FaaExternalCall ( &eci_parms ) ;
  Response ( "ECI ", sRetCode) ;

  return sRetCode ;

} /* end function ECICall */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: ECIRequestData                                              */
/*                                                                        */
/*  PURPOSE:  Call ECI to retrieve data after ECI has posted ASYNC_NOTIFY */
/*                                                                        */
/**************************************************************************/

SHORT ECIRequestData (void)

{

  int     Size ;
  SHORT   sRetCode ;

  Info ("Calling ECI to obtain commarea after ASYNC_NOTIFY.") ;

  /* clear the commarea */
  Size = sizeof ( szCommArea ) ;
  memset ( szCommArea , 0 , Size ) ;

  /* clear the ECI control block */
  memset ( &eci_parms,  0,  sizeof ( ECI_PARMS ) ) ;

  eci_parms.eci_call_type                   = ECI_GET_REPLY ;
  eci_parms.eci_commarea                    = szCommArea ;
  eci_parms.eci_commarea_length             = Size ;

  /* call the ECI */
  sRetCode = FaaExternalCall ( &eci_parms ) ;
  /* display the return code */
  Response ( "ECI ", sRetCode ) ;

  return sRetCode ;

} /* end function ECIRequestData */

/***************************************************************************/
/* DDE Server Window procedure                                             */
/***************************************************************************/
/***************************************************************************/
/*                                                                         */
/*  FUNCTION: DDEWindowProc                                                */
/*                                                                         */
/*  PURPOSE:  Handles all DDE messages received by the server application. */
/*                                                                         */
/***************************************************************************/

MRESULT EXPENTRY DDEWindowProc( HWND   hwnd,
                                ULONG  msg,
                                MPARAM mp1,
                                MPARAM mp2)

{
  MRESULT                lRetCode = 0L ;

  switch (msg)  {

    case WM_DDE_TERMINATE :
      ServerTerminate(hwnd, (HWND)mp1 ) ;
      break ;

    case WM_DDE_REQUEST :
      ServerRequest(hwnd, (HWND)mp1, mp2 ) ;
      break ;

    case ECI_MSG :

      /* display the ECI return code */
      Response ( "ECI Async Notify Msg ", SHORT2FROMMP ( mp1 ) ) ;

      if (ECIRequestData() != ECI_NO_ERROR) {
        /* the ECI call has failed. Copy an error message into the commarea */
        strcpy( szCommArea , "ECI Failure, check DDESERV2") ;
      } /* end if */

      Info ( "Commarea returned is  '%s'",  szCommArea ) ;

      /* now we need to reestablish Conversation details to send the data */
      /* back to the requesting client.                                   */

      if ( GetConvItem(hwnd)) {
        /* send data now and don't ask for acknowledgement */

        SendData(hwnd, GetHwndDDEClient(hwnd), szCommArea, TRUE ) ;
      } /* end if */

      break ;

    default :

      return (WinDefWindowProc(hwnd, msg, mp1, mp2));

    }  /* end switch */

    return lRetCode ;

}   /* end function DDEWindowProc */


/***************************************************************************/
/*                                                                         */
/*  FUNCTION: ServerInitiate                                               */
/*                                                                         */
/*  PURPOSE:  To initialise child window and send acknowledgement if OK    */
/*                                                                         */
/***************************************************************************/

void ServerInitiate (HWND hwndDDEClient, MPARAM mp2)
{

  PDDEINIT pddei = NULL ;          /* DDE initial structure ptr */

  /* address the DDE structure pointed to by mp2 */
  pddei = (PDDEINIT)mp2;

  /* check to see if the "application" named is ours */

  if (strcmp(THISAPPNAME, pddei->pszAppName) == 0 ) {

    Info ("DDE_INITIATE received for appl: %s Topic: %s",
          pddei->pszAppName, pddei->pszTopic);

    /* save the topic locally for now */

    strcpy(szTopicName, pddei->pszTopic) ;

    /************************************************************************/
    /* create the child window to handle the conversation                   */
    /************************************************************************/

    hwndDDEServer = WinCreateWindow(
                  hwndFrame,                     // parent window handle
                  "DDEServerClass",              // pointer to registered class name
                  NULL,                          // pointer to window text
                  0L,                            // window style
                  0,0,0,0,                       // position of window
                  hwndFrame,                     // owner-handle window
                  HWND_TOP,                      // handle to sibling window
                  ID_CLIENT,                     // window identifier
                  NULL,                          // pointer to buffer,class specific
                  NULL) ;                        // pointer to structure-presentation


    if ( hwndDDEServer == (HWND) NULL ) {
      WinMessageBox ( hwndFrame, hwndFrame,
                      "DDE Initialization failure",
                      THISAPPNAME, 0,
                      MB_ERROR|MB_OK ) ;

      return ;
    } /* endif */

    Info ("DDE Server window initialised OK, handle is %d" , hwndDDEServer ) ;

    if (AddConv(hwndDDEServer, hwndDDEClient, szTopicName )) {

      WinDdeRespond(hwndDDEClient, hwndDDEServer, THISAPPNAME, szTopicName,
                   (PCONVCONTEXT)NULL) ;

    }   /* end if */
    else {

      /* AddConv failed - at maximum conversation count - destroy the window */
      WinDestroyWindow(hwndDDEServer) ;
      Info ("Max No of clients attached. DDE Server window %d destroyed.", hwndDDEServer) ;
    } /* end else  */

  } /* end if */

  else {
    Info ("DDE_INITIATE received for appl: %s Topic: %s",
          pddei->pszAppName, pddei->pszTopic) ;

  } /* end else */

  return ;

} /* end function ServerInitiate */


/***************************************************************************/
/*                                                                         */
/*  FUNCTION: ServerTerminate                                              */
/*                                                                         */
/*  PURPOSE:  Terminates the DDE Conversation                              */
/*                                                                         */
/***************************************************************************/
void ServerTerminate(HWND hwndDDEServer, HWND hwndDDEClient)

{
  Info ("DDE_TERMINATE received from client %d.",  hwndDDEClient ) ;

  if (! IsConvInTerminateState( hwndDDEClient )) {

    /* Client has requested terminate: responsd with terminate */
    WinDdePostMsg(hwndDDEClient, hwndDDEServer, WM_DDE_TERMINATE, NULL, TRUE) ;

  } /* end if */

  RemoveConv ( hwndDDEServer ) ;
  WinDestroyWindow( hwndDDEServer ) ;
  Info ("DDE Server window %d destroyed.", hwndDDEServer) ;
  return ;

}  /* end function ServerTerminate */

/***************************************************************************/
/*                                                                         */
/*  FUNCTION: ServerRequest                                                */
/*                                                                         */
/*  PURPOSE:  Called when Server receives WM_DDE_REQUEST message.          */
/*            ECI is called (ASYNC_NOTIFY) passing the item as a commarea. */
/*                                                                         */
/***************************************************************************/

void ServerRequest(HWND hwndDDEServer, HWND hwndDDEClient, MPARAM mp2)
{

    PDDESTRUCT pddeIn  = NULL ;
    int        Size ;
    short      sRetCode ;

    /* function code */

    pddeIn = (PDDESTRUCT)mp2;              // address the supplied DDE structure

    if (!pddeIn->usFormat == CF_TEXT) {

      /* This appl supports text only - reject any other request with NACK */
      /* flip the fsStatus bit and return the same shared memory segment.  */

      pddeIn->fsStatus &= (~DDE_FACK);
      WinDdePostMsg(hwndDDEClient, hwndDDEServer, WM_DDE_ACK, pddeIn, TRUE);
      return ;

    } /* end if */

  /* Now do the ECI call using the 'topic' as the program name */
  /* and the 'item' as the contents of a commarea to be passed */

    if ( GetConvTopic(hwndDDEServer) ) {

      strcpy( szItemName, DDES_PSZITEMNAME(pddeIn)) ;
      strcpy( szCommArea, DDES_PSZITEMNAME(pddeIn)) ;

      sRetCode = ECICall ( szTopicName, hwndDDEServer ) ;

      if ( sRetCode != ECI_NO_ERROR ) {

        /* the ECI call has failed. Copy an error message into the  */
        /* return value and send as a normal DDE response. We must  */
        /* free the sent shared memory object as we create a new    */
        /* object to return the error message back to the client    */

        strcpy( szCommArea, "ECI Failure. Check Server.") ;
        SendData(hwndDDEServer, hwndDDEClient, szCommArea, TRUE) ;

      }  /* end if */

     /* ECI Call will post a message when it returns from CICS OS/2.*/
     /* To allow us to reply to client we must save the item which  */
     /* was requested in the conv structure so we can               */
     /* restore it when the ECI_MSG is received.                    */

     AddConvItem ( hwndDDEServer, szItemName) ;

    } /* end if */

  /* return and wait for ECI_MSG to be posted to this server.       */
  /* We must free the sent DDE structure at this stage as we will   */
  /* create a new one to reply to the client whatever happens next  */

  DosFreeMem ( pddeIn ) ;

  return ;

}   /* end function ServerRequest */

/***************************************************************************/
/*                                                                         */
/*  FUNCTION: SendData                                                     */
/*                                                                         */
/*  PURPOSE:  Send DDE data to the client.                                 */
/*                                                                         */
/***************************************************************************/
void SendData( HWND hwndDDEServer,
               HWND hwndDDEClient,
               CHAR * szItem,
               BOOL bRequestData )
{

  PDDESTRUCT  pddeOut = NULL;          /* ptr to dde data structure   */
  ULONG            rc = 0;             /* api return code holder      */
  PULONG pulSharedObj = NULL;          /* pointer to shared object    */
  PCHAR           psz = NULL;          /* string ptr for text format  */
  USHORT    cbObjSize = 0;             /* count of bytes to request   */
  USHORT       cbData = 0;             /* length of return data       */

  /**************************************************************************/
  /* Allocate a givable memory segment                                      */
  /**************************************************************************/

  cbObjSize = (USHORT)strlen(szItem)+ (USHORT) 1;
  rc = DosAllocSharedMem((PPVOID) &pulSharedObj, NULL, cbObjSize,
             PAG_COMMIT | PAG_READ | PAG_WRITE | OBJ_GIVEABLE);
  if (rc && pulSharedObj)
    /* error message is required in here */
    return ;

  /**************************************************************************/
  /* Fill in the new DDE structure                                          */
  /**************************************************************************/

  pddeOut = (PDDESTRUCT) pulSharedObj ;
  pddeOut->usFormat = CF_TEXT ;
  pddeOut->offszItemName = (USHORT)sizeof(DDESTRUCT) ; /* offset to item name*/

  strcpy(DDES_PSZITEMNAME(pddeOut), szItemName) ;      /* item name*/

  cbData = (USHORT)strlen(szItem)+(USHORT)1 ;
  pddeOut->cbData = cbData ;
  pddeOut->offabData = (USHORT)pddeOut->offszItemName + cbData ;
  pddeOut->fsStatus = (USHORT) 0 ;                /* set status flags*/

  psz = szItem;
  strcpy(DDES_PABDATA(pddeOut), psz) ;

  WinDdePostMsg(hwndDDEClient, hwndDDEServer, WM_DDE_DATA, pddeOut, TRUE) ;

  return ;

} /* end function send data */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: Terminate Converations                                      */
/*                                                                        */
/*  PURPOSE:  After WM_DESTROY, we must terminate all DDE converations    */
/*                                                                        */
/**************************************************************************/
void TerminateConversations( void )
{
  HWND  hwndDDEServer ;

  /* Send DDE_TERMINATE to each client in conversation with us */

  hwndDDEServer = (HWND) NULL ;

  while (hwndDDEServer = GetNextConv(hwndDDEServer)) {
    SendTerminate(hwndDDEServer, GetHwndDDEClient(hwndDDEServer)) ;
  } /* end while */

  return ;

} /* end function TerminateConversations */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: AddConv                                                     */
/*                                                                        */
/*  PURPOSE:  Register a new conversation with a client                   */
/*                                                                        */
/**************************************************************************/
BOOL AddConv(HWND hwndDDEServer, HWND hwndDDEClient, CHAR * szTopic )
{
  struct DDECONV * pConv ;

  if (nConvCount >= CONV_MAX_COUNT) {
    return (FALSE) ;
  } /* end if */

  if (FindConv(hwndDDEServer) != NULL) {
    /* conversation already added */
    return (FALSE) ;
  } /* end if */

  pConv = Conv + nConvCount++ ;
  pConv->hwndDDEServer = hwndDDEServer ;
  pConv->hwndDDEClient = hwndDDEClient ;
  strcpy(pConv->szTopic, szTopic) ;        /* copy the topic string */
  strcpy(pConv->szItem, "NULL") ;          /* item is currently null */
  pConv->bInClientRequestedTerminate = FALSE ;
  return (TRUE) ;

} /* end function AddConv */

/**************************************************************************/
/*                                                                        */
/*  FUNCTION: AddConvItem                                                 */
/*                                                                        */
/*  PURPOSE:  Update the conversation to contain the requested item       */
/*                                                                        */
/**************************************************************************/
BOOL AddConvItem ( HWND hwndDDEServer, CHAR * szItem )
{

  struct DDECONV * pConv ;

  if (!(pConv = FindConv(hwndDDEServer)))
    return FALSE ;

  strcpy(pConv->szItem, szItem) ;        /* copy the item string */
  return TRUE ;

} /* end function AddConvItem */

/**************************************************************************/
/*                                                                        */
/*  FUNCTION: GetConvTopic                                                */
/*                                                                        */
/*  PURPOSE:  Get the string representing the 'Topic' of the conversation */
/*            and copy it into global storage                             */
/*                                                                        */
/**************************************************************************/
BOOL GetConvTopic ( HWND hwndDDEServer)
{
  struct DDECONV * pConv ;

  if (!(pConv = FindConv(hwndDDEServer)))
    return FALSE ;

  strcpy(szTopicName, pConv->szTopic) ;
  return TRUE ;

} /* end function GetConvTopic */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: GetConvItem                                                 */
/*                                                                        */
/*  PURPOSE:  Get the string representing the 'item' of the conversation  */
/*            with the specified DDE Server Window and copy it into       */
/*            Global storage                                              */
/*                                                                        */
/**************************************************************************/
BOOL GetConvItem ( HWND hwndDDEServer )
{
  struct DDECONV * pConv ;

  if (!(pConv = FindConv(hwndDDEServer)))
    return FALSE ;

  strcpy(szItemName, pConv->szItem) ;
  return TRUE ;

} /* end function GetConvItem */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: AtLeastoneConvActive                                        */
/*                                                                        */
/*  PURPOSE:  Test whether any conversations are still active             */
/*                                                                        */
/**************************************************************************/
BOOL AtLeastOneConvActive( void )
{
    return (nConvCount? TRUE: FALSE);
}   /* end function AtLeastOneConvActive */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: FindConv                                                    */
/*                                                                        */
/*  PURPOSE:  Find Conversation for specified DDE server window           */
/*                                                                        */
/**************************************************************************/
struct DDECONV * NEAR FindConv( HWND hwndDDEServer)
{
  struct DDECONV *  pConv ;
  int               nConvIndex ;

  for (nConvIndex = 0, pConv = Conv ;
       nConvIndex < nConvCount ;
       nConvIndex++, pConv++) {

    if (pConv->hwndDDEServer == hwndDDEServer)
      return (pConv) ;
  } /* end for */

  return (NULL) ;

}   /* end function FindConv */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: GetNextConv                                                 */
/*                                                                        */
/*  PURPOSE:  Get the next client handle from the list of conversations   */
/*                                                                        */
/**************************************************************************/
HWND GetNextConv(HWND hwndDDEServer)
{
  struct DDECONV * pConv ;
  int           nConvIndex ;

  if (hwndDDEServer) {
    for (nConvIndex = 0, pConv = Conv ;
         nConvIndex < nConvCount ;
         nConvIndex++, pConv++) {

      if (pConv->hwndDDEServer == hwndDDEServer) {

        if (++nConvIndex < nConvCount)
          return (++pConv)->hwndDDEServer ;
        else
          return (HWND) NULL ;
      } /* end if */
    }  /* end for */
    return (HWND) NULL ;
  } /* end if */

  if (nConvCount > 0)
    return (Conv[0].hwndDDEServer) ;
  else
    return (HWND) NULL ;

}   /* end function GetNextConv */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: RemoveConv                                                  */
/*                                                                        */
/*  PURPOSE:  Remove the conversation from the list                       */
/*                                                                        */
/**************************************************************************/
void RemoveConv( HWND hwndDDEServer)
{
  struct DDECONV   * pConv ;
  int              nConvIndex ;

  for (nConvIndex = 0, pConv = Conv ;
       nConvIndex < nConvCount ;
       nConvIndex++, pConv++) {

    if (pConv->hwndDDEServer == hwndDDEServer)
      break ;
  } /* end for */

  nConvCount-- ;
  while (nConvIndex < nConvCount) {

    *pConv = *(pConv+1) ;
    nConvIndex++ ;
    pConv++ ;
  }  /* end while */

  return;

} /* end function RemoveConv */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: GetHwndDDEClient                                            */
/*                                                                        */
/*  PURPOSE:  Get the hwnd of the client in conversation with the         */
/*            specified DDE Server Window                                 */
/*                                                                        */
/**************************************************************************/
HWND GetHwndDDEClient( HWND hwndDDEServer )
{
  struct DDECONV * pConv ;

  if (!(pConv = FindConv(hwndDDEServer)))
    return (HWND) NULL ;
  return (pConv->hwndDDEClient) ;

} /* end function GetHwndDDEClient */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: SendTerminate                                               */
/*                                                                        */
/*  PURPOSE:  Post terminate message and set the terminating flag for     */
/*            the conversation                                            */
/*                                                                        */
/**************************************************************************/
void SendTerminate( HWND hwndDDEServer, HWND hwndDDEClient )
{

  SetConvInTerminateState(hwndDDEServer) ;

  WinDdePostMsg(hwndDDEClient, hwndDDEServer, WM_DDE_TERMINATE, NULL, TRUE) ;
  return ;

}

/**************************************************************************/
/*                                                                        */
/*  FUNCTION: SetConvInTerminateState                                     */
/*                                                                        */
/*  PURPOSE:  Set the specified conversations terminate state to true     */
/*                                                                        */
/**************************************************************************/
void SetConvInTerminateState( HWND hwndDDEServer )
{
  struct DDECONV * pConv ;

  if (pConv = FindConv(hwndDDEServer))
      pConv->bInClientRequestedTerminate = TRUE ;
  return;

} /* end function SetConvInTerminateState */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: IsConvInTerminateState                                      */
/*                                                                        */
/*  PURPOSE:  Test whether conversation with specified client is in       */
/*            process of being terminated                                 */
/*                                                                        */
/**************************************************************************/
BOOL IsConvInTerminateState( HWND hwndDDEServer )
{
  struct DDECONV * pConv;

  if (pConv = FindConv(hwndDDEServer))
    return (pConv->bInClientRequestedTerminate) ;
  else
    return (FALSE);

}   /* end function IsConvInTerminateState */

/***************************************************************************/
/*       Window output functions.                                          */
/***************************************************************************/
/* Info - write a line of information to the window                        */
/***************************************************************************/

void Info( char *Format, ... )
{
  va_list ArgPtr;

  va_start ( ArgPtr, Format ) ;
  vsprintf ( szBuffer, Format, ArgPtr ) ;
  va_end ( ArgPtr ) ;

  WinSendMsg ( hwndClient,
               LM_INSERTITEM,
               MPFROMSHORT ( LIT_END ),
               MPFROMP ( szBuffer ) ) ;
}

/***************************************************************************/
/* Response - display the response from an Eci Call                        */
/***************************************************************************/
void Response ( char *Call, short sRetCode )
{
   char *pszError ;

   switch ( sRetCode ) {
      case ECI_NO_ERROR:
         pszError = "ECI_NO_ERROR";
         break;
      case ECI_ERR_INVALID_DATA_LENGTH:
         pszError = "ECI_ERR_INVALID_DATA_LENGTH";
         break;
      case ECI_ERR_INVALID_EXTEND_MODE:
         pszError = "ECI_ERR_INVALID_EXTEND_MODE";
         break;
      case ECI_ERR_NO_CICS:
         pszError = "ECI_ERR_NO_CICS";
         break;
      case ECI_ERR_CICS_DIED:
         pszError = "ECI_ERR_CICS_DIED";
         break;
      case ECI_ERR_REQUEST_TIMEOUT:
         pszError = "ECI_ERR_REQUEST_TIMEOUT";
         break;
      case ECI_ERR_RESPONSE_TIMEOUT:
         pszError = "ECI_ERR_RESPONSE_TIMEOUT";
         break;
      case ECI_ERR_TRANSACTION_ABEND:
         pszError = "ECI_ERR_TRANSACTION_ABEND";
         break;
      case ECI_ERR_EXEC_NOT_RESIDENT:
         pszError = "ECI_ERR_EXEC_NOT_RESIDENT";
         break;
      case ECI_ERR_SYSTEM_ERROR:
         pszError = "ECI_ERR_SYSTEM_ERROR";
         break;
      case ECI_ERR_NULL_WIN_HANDLE:
         pszError = "ECI_ERR_NULL_WIN_HANDLE";
         break;
      case ECI_ERR_NULL_MESSAGE_ID:
         pszError = "ECI_ERR_NULL_MESSAGE_ID";
         break;
      case ECI_ERR_THREAD_CREATE_ERROR:
         pszError = "ECI_ERR_THREAD_CREATE_ERROR";
         break;
      case ECI_ERR_INVALID_CALL_TYPE:
         pszError = "ECI_ERR_INVALID_CALL_TYPE";
         break;
      case ECI_ERR_ALREADY_ACTIVE:
         pszError = "ECI_ERR_ALREADY_ACTIVE";
         break;
      case ECI_ERR_RESOURCE_SHORTAGE:
         pszError = "ECI_ERR_RESOURCE_SHORTAGE";
         break;
      case ECI_ERR_NO_SESSIONS:
         pszError = "ECI_ERR_NO_SESSIONS";
         break;
      case ECI_ERR_NULL_SEM_HANDLE:
         pszError = "ECI_ERR_NULL_SEM_HANDLE";
         break;
      case ECI_ERR_INVALID_DATA_AREA:
         pszError = "ECI_ERR_INVALID_DATA_AREA";
         break;
      default:
         pszError = "!!!Unknown!!!";
         break;
   } /* endswitch */

   Info ( "%s returned: %s", Call, pszError ) ;

   if ( sRetCode == ECI_ERR_SYSTEM_ERROR ) {
      /* Display the system error response: */
      Info ( "SysErr returns: %d(0x%04.4X)",
             eci_parms.eci_sys_return_code,
             eci_parms.eci_sys_return_code ) ;
   } /* endif */
}
